@@ -1,6 +1,6 @@
 import contextlib
 from datetime import datetime
-from typing import Callable, ContextManager, Dict, Iterator, List, Optional, Union
+from typing import Any, Callable, ContextManager, Dict, List, Optional, Union
 
 import numpy as np
 import pandas as pd
@@ -12,8 +12,10 @@ from six import reraise
 from feast import FeatureView
 from feast.data_source import DataSource
 from feast.errors import InvalidEntityType
+from feast.feature_view import DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL
 from feast.infra.offline_stores import offline_utils
 from feast.infra.offline_stores.offline_store import OfflineStore, RetrievalJob
+from feast.on_demand_feature_view import OnDemandFeatureView
 from feast.registry import Registry
 from feast.repo_config import FeastConfigBaseModel, RepoConfig
 from feast_hive.hive_source import HiveSource
@@ -31,6 +33,12 @@ except ImportError as e:
     raise FeastExtrasDependencyImportError("hive", str(e))
 
 
+# default Hive settings that can be overridden in offline_store:hive_conf of feature_store.yaml
+DEFAULT_HIVE_CONF = {
+    "hive.strict.checks.cartesian.product": "false",
+}
+
+
 class HiveOfflineStoreConfig(FeastConfigBaseModel):
     """ Offline store config for Hive """
 
@@ -79,6 +87,16 @@ class HiveOfflineStoreConfig(FeastConfigBaseModel):
     kerberos_service_name: StrictStr = "impala"
     """ Specify particular impalad service principal. """
 
+    def __init__(self, **data: Any):
+        if "hive_conf" not in data:
+            data["hive_conf"] = DEFAULT_HIVE_CONF.copy()
+        elif type(data["hive_conf"]) == dict:
+            data["hive_conf"] = dict(list(DEFAULT_HIVE_CONF.items()) + list(data["hive_conf"].items()))
+        else:
+            raise TypeError("hive_conf should be a dictionary!")
+
+        super().__init__(**data)
+
 
 class HiveConnection:
     def __init__(self, store_config: HiveOfflineStoreConfig):
@@ -143,19 +161,24 @@ class HiveOfflineStore(OfflineStore):
         start_date = _format_datetime(start_date)
         end_date = _format_datetime(end_date)
 
-        query = f"""
-                SELECT {field_string}
-                FROM (
-                    SELECT {field_string},
-                    ROW_NUMBER() OVER({partition_by_join_key_string} ORDER BY {timestamp_desc_string}) AS feast_row_
-                    FROM {from_expression} t1
-                    WHERE {event_timestamp_column} BETWEEN TIMESTAMP('{start_date}') AND TIMESTAMP('{end_date}')
-                ) t2
-                WHERE feast_row_ = 1
-                """
+        queries = [
+            "SET hive.resultset.use.unique.column.names=false",
+            f"""
+            SELECT 
+                {field_string}
+                {f", {repr(DUMMY_ENTITY_VAL)} AS {DUMMY_ENTITY_ID}" if not join_key_columns else ""}
+            FROM (
+                SELECT {field_string},
+                ROW_NUMBER() OVER({partition_by_join_key_string} ORDER BY {timestamp_desc_string}) AS feast_row_
+                FROM {from_expression} t1
+                WHERE {event_timestamp_column} BETWEEN TIMESTAMP('{start_date}') AND TIMESTAMP('{end_date}')
+            ) t2
+            WHERE feast_row_ = 1
+            """,
+        ]
 
         conn = HiveConnection(config.offline_store)
-        return HiveRetrievalJob(conn, query)
+        return HiveRetrievalJob(conn, queries)
 
     @staticmethod
     def get_historical_features(
@@ -171,7 +194,7 @@ class HiveOfflineStore(OfflineStore):
         conn = HiveConnection(config.offline_store)
 
         @contextlib.contextmanager
-        def query_generator() -> Iterator[List[str]]:
+        def query_generator() -> ContextManager[List[str]]:
             table_name = offline_utils.get_temp_entity_table_name()
 
             try:
@@ -191,14 +214,15 @@ class HiveOfflineStore(OfflineStore):
                     entity_schema, expected_join_keys, entity_df_event_timestamp_col
                 )
 
-                query_context = offline_utils.get_feature_view_query_context(
+                query_contexts = offline_utils.get_feature_view_query_context(
                     feature_refs, feature_views, registry, project,
                 )
 
                 rendered_query = offline_utils.build_point_in_time_query(
-                    query_context,
+                    query_contexts,
                     left_table_query_string=table_name,
                     entity_df_event_timestamp_col=entity_df_event_timestamp_col,
+                    # entity_df_columns=entity_schema.keys(), # FIXME: uncomment this for <master> branch of feast repo
                     query_template=MULTIPLE_FEATURE_VIEW_POINT_IN_TIME_JOIN,
                     full_feature_names=full_feature_names,
                 )
@@ -206,7 +230,8 @@ class HiveOfflineStore(OfflineStore):
                 # In order to use `REGEX Column Specification`, need set `hive.support.quoted.identifiers` to None.
                 # Can study more here: https://cwiki.apache.org/confluence/display/Hive/LanguageManual+Select
                 queries = [
-                    "SET hive.support.quoted.identifiers=None",
+                    "SET hive.support.quoted.identifiers=none",
+                    "SET hive.resultset.use.unique.column.names=false",
                     "SET hive.exec.temporary.table.storage=memory",
                 ] + rendered_query.split(";")
 
@@ -216,7 +241,14 @@ class HiveOfflineStore(OfflineStore):
                 with conn.cursor() as cursor:
                     cursor.execute(f"DROP TABLE IF EXISTS {table_name}")
 
-        return HiveRetrievalJob(conn, query_generator)
+        return HiveRetrievalJob(
+            conn,
+            query_generator,
+            full_feature_names=full_feature_names,
+            on_demand_feature_views=OnDemandFeatureView.get_requested_odfvs(
+                feature_refs, project, registry
+            ),
+        )
 
 
 class HiveRetrievalJob(RetrievalJob):
@@ -224,6 +256,8 @@ class HiveRetrievalJob(RetrievalJob):
         self,
         conn: HiveConnection,
         queries: Union[str, List[str], Callable[[], ContextManager[List[str]]]],
+        full_feature_names: bool = False,
+        on_demand_feature_views: Optional[List[OnDemandFeatureView]] = None,
     ):
         assert (
             isinstance(queries, str) or isinstance(queries, list) or callable(queries)
@@ -234,19 +268,34 @@ class HiveRetrievalJob(RetrievalJob):
         else:
             if isinstance(queries, str):
                 queries = [queries]
+            elif not isinstance(queries, list):
+                raise TypeError(
+                    "queries should be a context manager yielding list[queries]"
+                    "or list[str] or str"
+                )
 
             @contextlib.contextmanager
-            def query_generator() -> Iterator[List[str]]:
+            def query_generator() -> ContextManager[List[str]]:
                 yield queries
 
             self._queries_generator = query_generator
 
         self._conn = conn
+        self._full_feature_names = full_feature_names
+        self._on_demand_feature_views = on_demand_feature_views
 
-    def to_df(self) -> pd.DataFrame:
-        return self.to_arrow().to_pandas()
+    @property
+    def full_feature_names(self) -> bool:
+        return self._full_feature_names
 
-    def to_arrow(self) -> pa.Table:
+    @property
+    def on_demand_feature_views(self) -> Optional[List[OnDemandFeatureView]]:
+        return self._on_demand_feature_views
+
+    def _to_df_internal(self) -> pd.DataFrame:
+        return self._to_arrow_internal().to_pandas()
+
+    def _to_arrow_internal(self) -> pa.Table:
         with self._queries_generator() as queries:
             with self._conn.cursor() as cursor:
                 for query in queries:
